"""
MViT v2, adapted from
https://github.com/facebookresearch/mvit
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/mvitv2.py

"MViTv2: Improved Multiscale Vision Transformers for Classification and Detection",
https://arxiv.org/abs/2112.01526

Changes from original:
* Simplified MultiScaleAttention - most options removed
"""

# Reference license: Apache-2.0 (both)

import math
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


def pre_pool(
    x: torch.Tensor, hw_shape: tuple[int, int], has_cls_token: bool
) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
    (H, W) = hw_shape
    if has_cls_token is True:
        cls_tok = x[:, :, :1, :]
        x = x[:, :, 1:, :]
    else:
        cls_tok = None

    x = x.reshape(-1, H, W, x.size(-1)).permute(0, 3, 1, 2).contiguous()

    return (x, cls_tok)


def post_pool(x: torch.Tensor, num_heads: int, cls_tok: Optional[torch.Tensor]) -> tuple[torch.Tensor, tuple[int, int]]:
    hw_shape = (x.size(2), x.size(3))
    l_pooled = x.size(2) * x.size(3)
    x = x.reshape(-1, num_heads, x.shape[1], l_pooled).transpose(2, 3)
    if cls_tok is not None:
        x = torch.concat((cls_tok, x), dim=2)

    return (x, hw_shape)


def cal_rel_pos_spatial(
    attn: torch.Tensor,
    q: torch.Tensor,
    has_cls_token: bool,
    q_shape: tuple[int, int],
    k_shape: tuple[int, int],
    rel_pos_h: torch.Tensor,
    rel_pos_w: torch.Tensor,
) -> torch.Tensor:
    sp_idx = 1 if has_cls_token is True else 0
    (q_h, q_w) = q_shape
    (k_h, k_w) = k_shape

    # Scale up rel pos if shapes for q and k are different.
    q_h_ratio = max(k_h / q_h, 1.0)
    k_h_ratio = max(q_h / k_h, 1.0)
    dist_h = (
        torch.arange(q_h, device=q.device).unsqueeze(-1) * q_h_ratio
        - torch.arange(k_h, device=q.device).unsqueeze(0) * k_h_ratio
    )
    dist_h += (k_h - 1) * k_h_ratio
    q_w_ratio = max(k_w / q_w, 1.0)
    k_w_ratio = max(q_w / k_w, 1.0)
    dist_w = (
        torch.arange(q_w, device=q.device).unsqueeze(-1) * q_w_ratio
        - torch.arange(k_w, device=q.device).unsqueeze(0) * k_w_ratio
    )
    dist_w += (k_w - 1) * k_w_ratio

    rel_h = rel_pos_h[dist_h.long()]
    rel_w = rel_pos_w[dist_w.long()]

    (B, n_head, _, dim) = q.shape

    r_q = q[:, :, sp_idx:].reshape(B, n_head, q_h, q_w, dim)
    rel_h = torch.einsum("byhwc,hkc->byhwk", r_q, rel_h)
    rel_w = torch.einsum("byhwc,wkc->byhwk", r_q, rel_w)

    attn[:, :, sp_idx:, sp_idx:] = (
        attn[:, :, sp_idx:, sp_idx:].view(B, -1, q_h, q_w, k_h, k_w) + rel_h.unsqueeze(-1) + rel_w.unsqueeze(-2)
    ).view(B, -1, q_h * q_w, k_h * k_w)

    return attn


class SequentialWithShape(nn.Sequential):
    def forward(  # pylint: disable=arguments-differ
        self, x: torch.Tensor, hw_shape: tuple[int, int]
    ) -> tuple[torch.Tensor, tuple[int, int]]:
        for module in self:
            (x, hw_shape) = module(x, hw_shape)

        return (x, hw_shape)


class PatchEmbed(nn.Module):
    def __init__(
        self, dim_in: int, dim_out: int, kernel: tuple[int, int], stride: tuple[int, int], padding: tuple[int, int]
    ) -> None:
        super().__init__()

        self.proj = nn.Conv2d(
            dim_in,
            dim_out,
            kernel_size=kernel,
            stride=stride,
            padding=padding,
        )

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, tuple[int, int]]:
        x = self.proj(x)
        (H, W) = x.shape[2:4]

        x = x.flatten(2).transpose(1, 2)

        return (x, (H, W))


class MultiScaleAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        input_size: tuple[int, int],
        num_heads: int,
        qkv_bias: bool,
        kernel_q: tuple[int, int],
        kernel_kv: tuple[int, int],
        stride_q: tuple[int, int],
        stride_kv: tuple[int, int],
        has_cls_token: bool,
    ) -> None:
        super().__init__()

        self.num_heads = num_heads
        self.dim_out = dim_out
        self.stride_q = stride_q
        self.stride_kv = stride_kv
        head_dim = dim_out // num_heads
        self.scale = head_dim**-0.5
        self.has_cls_token = has_cls_token
        padding_q = [int(q // 2) for q in kernel_q]
        padding_kv = [int(kv // 2) for kv in kernel_kv]

        self.qkv = nn.Linear(dim, dim_out * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim_out, dim_out)

        dim_conv = dim_out // num_heads

        # Skip pooling with kernel and stride size of (1, 1)
        if math.prod(kernel_q) != 1 or math.prod(stride_q) != 1:
            self.pool_q = nn.Conv2d(
                dim_conv,
                dim_conv,
                kernel_q,
                stride=stride_q,
                padding=padding_q,
                groups=dim_conv,
                bias=False,
            )
            self.norm_q = nn.LayerNorm(dim_conv)
        else:
            self.pool_q = None
            self.norm_q = None

        if math.prod(kernel_kv) != 1 or math.prod(stride_kv) != 1:
            self.pool_k = nn.Conv2d(
                dim_conv,
                dim_conv,
                kernel_kv,
                stride=stride_kv,
                padding=padding_kv,
                groups=dim_conv,
                bias=False,
            )
            self.norm_k = nn.LayerNorm(dim_conv)

            self.pool_v = nn.Conv2d(
                dim_conv,
                dim_conv,
                kernel_kv,
                stride=stride_kv,
                padding=padding_kv,
                groups=dim_conv,
                bias=False,
            )
            self.norm_v = nn.LayerNorm(dim_conv)
        else:
            self.pool_k = None
            self.norm_k = None
            self.pool_v = None
            self.norm_v = None

        # Relative pos embedding
        q_size_h = input_size[0] // stride_q[0]
        q_size_w = input_size[1] // stride_q[1]
        kv_size_h = input_size[0] // stride_kv[0]
        kv_size_w = input_size[1] // stride_kv[1]
        rel_sp_dim_h = 2 * max(q_size_h, kv_size_h) - 1
        rel_sp_dim_w = 2 * max(q_size_w, kv_size_w) - 1

        self.rel_pos_h = nn.Parameter(torch.zeros(rel_sp_dim_h, head_dim))
        self.rel_pos_w = nn.Parameter(torch.zeros(rel_sp_dim_w, head_dim))

        # Weights initialization
        nn.init.trunc_normal_(self.rel_pos_h, std=0.02)
        nn.init.trunc_normal_(self.rel_pos_w, std=0.02)

    def forward(self, x: torch.Tensor, hw_shape: tuple[int, int]) -> tuple[torch.Tensor, tuple[int, int]]:
        (B, N, _) = x.size()

        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, -1).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(dim=0)

        if self.pool_q is not None:
            (q, q_tok) = pre_pool(q, hw_shape, self.has_cls_token)
            q = self.pool_q(q)
            (q, q_shape) = post_pool(q, self.num_heads, q_tok)
            q = self.norm_q(q)
        else:
            q_shape = hw_shape

        if self.pool_k is not None:
            (k, k_tok) = pre_pool(k, hw_shape, self.has_cls_token)
            k = self.pool_k(k)
            (k, k_shape) = post_pool(k, self.num_heads, k_tok)
            k = self.norm_k(k)
        else:
            k_shape = hw_shape

        if self.pool_v is not None:
            (v, v_tok) = pre_pool(v, hw_shape, self.has_cls_token)
            v = self.pool_v(v)
            (v, _) = post_pool(v, self.num_heads, v_tok)
            v = self.norm_v(v)

        attn = (q * self.scale) @ k.transpose(-2, -1)
        attn = cal_rel_pos_spatial(attn, q, self.has_cls_token, q_shape, k_shape, self.rel_pos_h, self.rel_pos_w)

        attn = attn.softmax(dim=-1)
        x = attn @ v
        if self.has_cls_token is True:
            x[:, :, 1:, :] = x[:, :, 1:, :] + q[:, :, 1:, :]
        else:
            x = x + q

        x = x.transpose(1, 2).reshape(B, -1, self.dim_out)
        x = self.proj(x)

        return (x, q_shape)


class MultiScaleBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        num_heads: int,
        input_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        kernel_q: tuple[int, int],
        kernel_kv: tuple[int, int],
        stride_q: tuple[int, int],
        stride_kv: tuple[int, int],
        has_cls_token: bool,
        dim_mul_in_att: bool,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out
        self.num_heads = num_heads
        self.norm1 = nn.LayerNorm(dim)
        self.has_cls_token = has_cls_token
        self.dim_mul_in_att = dim_mul_in_att

        att_dim = dim_out if dim_mul_in_att is True else dim
        self.attn = MultiScaleAttention(
            dim,
            att_dim,
            input_size=input_size,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            kernel_q=kernel_q,
            kernel_kv=kernel_kv,
            stride_q=stride_q,
            stride_kv=stride_kv,
            has_cls_token=has_cls_token,
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(att_dim)
        self.mlp = MLP(att_dim, [int(att_dim * mlp_ratio), dim_out], activation_layer=nn.GELU, inplace=None)

        if self.dim_mul_in_att is True and self.dim != self.dim_out:
            self.proj_attn = nn.Linear(dim, dim_out)
        else:
            self.proj_attn = None

        if self.dim_mul_in_att is False and self.dim != self.dim_out:
            self.proj_mlp = nn.Linear(dim, dim_out)
        else:
            self.proj_mlp = None

        if len(stride_q) > 0 and math.prod(stride_q) > 1:
            kernel_skip = [s + 1 if s > 1 else s for s in stride_q]
            stride_skip = stride_q
            padding_skip = [int(skip // 2) for skip in kernel_skip]
            self.pool_skip = nn.MaxPool2d(kernel_skip, stride_skip, padding_skip, ceil_mode=False)
        else:
            self.pool_skip = None

    def _shortcut_pool(self, x: torch.Tensor, hw_shape: tuple[int, int]) -> torch.Tensor:
        if self.has_cls_token is True:
            cls_tok = x[:, :1, :]
            x = x[:, 1:, :]
        else:
            cls_tok = None

        (B, _, C) = x.size()
        (H, W) = hw_shape
        x = x.reshape(B, H, W, C).permute(0, 3, 1, 2).contiguous()
        x = self.pool_skip(x)
        x = x.reshape(B, C, -1).transpose(1, 2)
        if cls_tok is not None:
            x = torch.concat((cls_tok, x), dim=1)

        return x

    def forward(self, x: torch.Tensor, hw_shape: tuple[int, int]) -> tuple[torch.Tensor, tuple[int, int]]:
        x_norm = self.norm1(x)
        (x_block, hw_shape_new) = self.attn(x_norm, hw_shape)

        if self.proj_attn is not None:
            x = self.proj_attn(x_norm)

        if self.pool_skip is not None:
            x_res = self._shortcut_pool(x, hw_shape)
        else:
            x_res = x

        x = x_res + self.drop_path(x_block)
        x_norm = self.norm2(x)
        x_mlp = self.mlp(x_norm)

        if self.proj_mlp is not None:
            x = self.proj_mlp(x_norm)

        x = x + self.drop_path(x_mlp)

        return (x, hw_shape_new)


class MultiScaleVitStage(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        num_heads: int,
        input_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        kernel_q: tuple[int, int],
        kernel_kv: tuple[int, int],
        stride_q: tuple[int, int],
        stride_kv: tuple[int, int],
        has_cls_token: bool,
        dim_mul_in_att: bool,
        drop_path: list[float],
    ) -> None:
        super().__init__()

        self.blocks = nn.ModuleList()
        if dim_mul_in_att is True:
            out_dims = (dim_out,) * depth
        else:
            out_dims = (dim,) * (depth - 1) + (dim_out,)

        for i in range(depth):
            self.blocks.append(
                MultiScaleBlock(
                    dim=dim,
                    dim_out=out_dims[i],
                    num_heads=num_heads,
                    input_size=input_size,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    kernel_q=kernel_q,
                    kernel_kv=kernel_kv,
                    stride_q=stride_q if i == 0 else (1, 1),
                    stride_kv=stride_kv,
                    has_cls_token=has_cls_token,
                    dim_mul_in_att=dim_mul_in_att,
                    drop_path=drop_path[i],
                )
            )
            dim = out_dims[i]
            if i == 0:
                input_size = (input_size[0] // stride_q[0], input_size[1] // stride_q[1])

    def forward(self, x: torch.Tensor, hw_shape: tuple[int, int]) -> tuple[torch.Tensor, tuple[int, int]]:
        for blk in self.blocks:
            (x, hw_shape) = blk(x, hw_shape)

        return (x, hw_shape)


# pylint: disable=invalid-name
class MViT_v2(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        img_size = self.size
        stride_q = [(1, 1), (2, 2), (2, 2), (2, 2)]
        kernel_qkv = (3, 3)
        _stride_kv = (4, 4)
        depths: list[int] = self.config["depths"]
        embed_dim: int = self.config["embed_dim"]
        base_heads: int = self.config["base_heads"]
        dim_mul_in_att: bool = self.config["dim_mul_in_att"]
        use_cls_token: bool = self.config["use_cls_token"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stride_q = stride_q
        num_stages = len(depths)
        embed_dims = tuple(embed_dim * 2**i for i in range(num_stages))
        num_heads = tuple(base_heads * 2**i for i in range(num_stages))

        stride_kv = []
        for i in range(num_stages):
            if min(stride_q[i]) > 1:
                _stride_kv = (max(_stride_kv[0] // stride_q[i][0], 1), max(_stride_kv[1] // stride_q[i][1], 1))

            stride_kv.append(_stride_kv)

        self.patch_embed = PatchEmbed(
            dim_in=self.input_channels, dim_out=embed_dims[0], kernel=(7, 7), stride=(4, 4), padding=(3, 3)
        )

        if use_cls_token is True:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        else:
            self.cls_token = None

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        input_size = (img_size[0] // 4, img_size[1] // 4)
        for i in range(num_stages):
            if dim_mul_in_att is True:
                dim_out = embed_dims[i]
            else:
                dim_out = embed_dims[min(i + 1, num_stages - 1)]

            stage = MultiScaleVitStage(
                dim=embed_dim,
                dim_out=dim_out,
                depth=depths[i],
                num_heads=num_heads[i],
                input_size=input_size,
                mlp_ratio=4.0,
                qkv_bias=True,
                kernel_q=kernel_qkv,
                kernel_kv=kernel_qkv,
                stride_q=stride_q[i],
                stride_kv=stride_kv[i],
                has_cls_token=use_cls_token,
                dim_mul_in_att=dim_mul_in_att,
                drop_path=dpr[i],
            )
            stages[f"stage{i+1}"] = stage
            return_channels.append(dim_out)
            embed_dim = dim_out
            input_size = (input_size[0] // stride_q[i][0], input_size[1] // stride_q[i][1])

        self.body = SequentialWithShape(stages)
        self.norm = nn.LayerNorm(embed_dim)
        self.return_channels = return_channels
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = embed_dims[0]
        self.encoding_size = embed_dim

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (x, hw_shape) = self.patch_embed(x)
        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(x.size(0), -1, -1)
            x = torch.concat((cls_tokens, x), dim=1)

        out = {}
        for name, module in self.body.named_children():
            (x, hw_shape) = module(x, hw_shape)
            if name in self.return_stages:
                x_inter = x
                if self.cls_token is not None:
                    x_inter = x_inter[:, 1:]

                x_inter = x_inter.reshape(x.size(0), hw_shape[0], hw_shape[1], -1).permute(0, 3, 1, 2)
                out[name] = x_inter

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.patch_embed.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        B = x.size(0)

        (x, hw_shape) = self.patch_embed(x)
        x = mask_tensor(
            x.permute(0, 2, 1).reshape(B, -1, hw_shape[0], hw_shape[1]),
            mask,
            patch_factor=self.max_stride // self.stem_stride,
            mask_token=mask_token,
        )
        x = x.flatten(2).transpose(1, 2)

        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(B, -1, -1)
            x = torch.concat((cls_tokens, x), dim=1)

        (x, _) = self.body(x, hw_shape)
        x = self.norm(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            if self.cls_token is not None:
                features = x[:, 1:]
            else:
                features = x

            features = features.reshape(B, hw_shape[0], hw_shape[1], -1).permute(0, 3, 1, 2)
            result["features"] = features

        if return_keys in ("all", "embedding"):
            if self.cls_token is not None:
                result["embedding"] = x[:, 0]
            else:
                result["embedding"] = x.mean(1)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        (x, hw_shape) = self.patch_embed(x)
        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(x.size(0), -1, -1)
            x = torch.concat((cls_tokens, x), dim=1)

        (x, _) = self.body(x, hw_shape)
        x = self.norm(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)

        if self.cls_token is not None:
            x = x[:, 0]
        else:
            x = x.mean(1)

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        input_size = (new_size[0] // 4, new_size[1] // 4)
        for i, module in enumerate(self.body.children()):
            if isinstance(module, MultiScaleVitStage):
                idx = 0
                for m in module.modules():
                    if isinstance(m, MultiScaleBlock):
                        q_size_h = input_size[0] // m.attn.stride_q[0]
                        q_size_w = input_size[1] // m.attn.stride_q[1]
                        kv_size_h = input_size[0] // m.attn.stride_kv[0]
                        kv_size_w = input_size[1] // m.attn.stride_kv[1]
                        rel_sp_dim_h = 2 * max(q_size_h, kv_size_h) - 1
                        rel_sp_dim_w = 2 * max(q_size_w, kv_size_w) - 1

                        rel_pos_h = m.attn.rel_pos_h
                        rel_pos_h_resized = F.interpolate(
                            rel_pos_h.reshape(1, rel_pos_h.shape[0], -1).permute(0, 2, 1),
                            size=rel_sp_dim_h,
                            mode="linear",
                        )
                        rel_pos_w = m.attn.rel_pos_w
                        rel_pos_w_resized = F.interpolate(
                            rel_pos_w.reshape(1, rel_pos_w.shape[0], -1).permute(0, 2, 1),
                            size=rel_sp_dim_w,
                            mode="linear",
                        )

                        m.attn.rel_pos_h = nn.Parameter(rel_pos_h_resized.reshape(-1, rel_sp_dim_h).permute(1, 0))
                        m.attn.rel_pos_w = nn.Parameter(rel_pos_w_resized.reshape(-1, rel_sp_dim_w).permute(1, 0))

                        if idx == 0:
                            input_size = (input_size[0] // self.stride_q[i][0], input_size[1] // self.stride_q[i][1])

                        idx += 1


registry.register_model_config(
    "mvit_v2_t",
    MViT_v2,
    config={
        "depths": [1, 2, 5, 2],
        "embed_dim": 96,
        "base_heads": 1,
        "dim_mul_in_att": True,
        "use_cls_token": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "mvit_v2_s",
    MViT_v2,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "base_heads": 1,
        "dim_mul_in_att": True,
        "use_cls_token": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "mvit_v2_b",
    MViT_v2,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "base_heads": 1,
        "dim_mul_in_att": True,
        "use_cls_token": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "mvit_v2_l",
    MViT_v2,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "base_heads": 2,
        "dim_mul_in_att": False,
        "use_cls_token": False,
        "drop_path_rate": 0.5,
    },
)
registry.register_model_config(
    "mvit_v2_t_cls",
    MViT_v2,
    config={
        "depths": [1, 2, 5, 2],
        "embed_dim": 96,
        "base_heads": 1,
        "dim_mul_in_att": True,
        "use_cls_token": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "mvit_v2_s_cls",
    MViT_v2,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "base_heads": 1,
        "dim_mul_in_att": True,
        "use_cls_token": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "mvit_v2_b_cls",
    MViT_v2,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "base_heads": 1,
        "dim_mul_in_att": True,
        "use_cls_token": True,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "mvit_v2_l_cls",
    MViT_v2,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "base_heads": 2,
        "dim_mul_in_att": True,
        "use_cls_token": True,
        "drop_path_rate": 0.5,
    },
)
registry.register_model_config(
    "mvit_v2_h_cls",
    MViT_v2,
    config={
        "depths": [4, 8, 60, 8],
        "embed_dim": 192,
        "base_heads": 3,
        "dim_mul_in_att": True,
        "use_cls_token": True,
        "drop_path_rate": 0.8,
    },
)

registry.register_weights(
    "mvit_v2_t_il-all",
    {
        "url": "https://huggingface.co/birder-project/mvit_v2_t_il-all/resolve/main",
        "description": "MViT v2 tiny model trained on the il-all dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 91.2,
                "sha256": "2f15635de9a9020b8e7123d3a342c072212fa720e65a5cae910efd202d427298",
            }
        },
        "net": {"network": "mvit_v2_t", "tag": "il-all"},
    },
)



"""
Hiera, adapted from
https://github.com/facebookresearch/hiera/blob/main/hiera/hiera.py

Paper "Hiera: A Hierarchical Vision Transformer without the Bells-and-Whistles", https://arxiv.org/abs/2306.00989

AbsWin variant from the paper "Window Attention is Bugged: How not to Interpolate Position Embeddings",
https://arxiv.org/abs/2311.05613

Changes from original:
* Support only 2d
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_from_indices
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.vit import adjust_position_embedding


def get_resized_mask(target_size: list[int], mask: torch.Tensor) -> torch.Tensor:
    # target_size: [(T), (H), W]
    # (spatial) mask: [B, C, (t), (h), w]
    if mask is None:
        return mask

    assert len(mask.shape[2:]) == len(target_size)
    if mask.shape[2:] != target_size:
        return F.interpolate(mask.float(), size=target_size)

    return mask


def undo_windowing(x: torch.Tensor, shape: tuple[int, int], mu_shape: list[int]) -> torch.Tensor:
    """
    Restore spatial organization by undoing windowed organization of mask units.

    Parameters
    ----------
    x
        Organized by mask units windows, e.g. in 2d [B, #MUy*#MUx, MUy, MUx, C]
    shape
        Current spatial shape, if it were not organized into mask unit windows, e.g. in 2d [B, #MUy*MUy, #MUx*MUx, C]
    mu_shape
        current mask unit shape, e.g. in 2d [MUy, MUx]

    Returns
    -------
    Tensor with shape of [B, #MUy*MUy, #MUx*MUx, C]
    """

    D = len(shape)
    (B, C) = x.shape[0], x.shape[-1]
    # [B, #MUy*#MUx, MUy, MUx, C] -> [B, #MUy, #MUx, MUy, MUx, C]
    num_mu = [s // mu for s, mu in zip(shape, mu_shape)]
    x = x.view(B, *num_mu, *mu_shape, C)

    # [B, #MUy, #MUx, MUy, MUx, C] -> [B, #MUy*MUy, #MUx*MUx, C]
    permute = [0] + sum([list(p) for p in zip(range(1, 1 + D), range(1 + D, 1 + 2 * D))], []) + [len(x.shape) - 1]
    x = x.permute(permute).reshape(B, *shape, C)

    return x


class PatchEmbed(nn.Module):
    def __init__(
        self, dim_in: int, dim_out: int, kernel: tuple[int, int], stride: tuple[int, int], padding: tuple[int, int]
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(dim_in, dim_out, kernel_size=kernel, stride=stride, padding=padding)

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        if mask is not None:
            mask = get_resized_mask(target_size=x.shape[2:], mask=mask)
            x = self.proj(x * mask.to(torch.bool))
        else:
            x = self.proj(x)

        x = x.reshape(x.shape[0], x.shape[1], -1).transpose(2, 1)

        return x


class Unroll(nn.Module):
    """
    Reorders the tokens such that patches are contiguous in memory.
    E.g., given [B, (H, W), C] and stride of (Sy, Sx), this will re-order the tokens as
                           [B, (Sy, Sx, H // Sy, W // Sx), C]

    This allows operations like Max2d to be computed as x.view(B, Sx*Sy, -1, C).max(dim=1).
    Not only is this faster, but it also makes it easy to support inputs of arbitrary
    dimensions in addition to patch-wise sparsity.

    Performing this operation multiple times in sequence puts entire windows as contiguous
    in memory. For instance, if you applied the stride (2, 2) 3 times, entire windows of
    size 8x8 would be contiguous in memory, allowing operations like mask unit attention
    computed easily and efficiently, while also allowing max to be applied sequentially.

    Note: This means that intermediate values of the model are not in HxW order, so they
    need to be re-rolled if you want to use the intermediate values as a HxW feature map.
    The last block of the network is fine though, since by then the strides are all consumed.

    Modified to only fit 2d due to TorchScript constraints.
    """

    def __init__(
        self, input_size: tuple[int, int], patch_stride: tuple[int, int], unroll_schedule: list[tuple[int, int]]
    ) -> None:
        super().__init__()
        self.size = (input_size[0] // patch_stride[0], input_size[1] // patch_stride[1])
        self.schedule = unroll_schedule

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Input: flattened patch embeddings [B, N, C]
        Output: patch embeddings [B, N, C] permuted such that [B, 4, N//4, C].max(1) etc. performs MaxPoolNd
        """

        (B, _, C) = x.shape
        cur_size = self.size
        x = x.view(B, *cur_size, C)

        for strides in self.schedule:
            # Move patches with the given strides to the batch dimension

            # Create a view of the tensor with the patch stride as separate dims
            # For example in 2d: [B, H // Sy, Sy, W // Sx, Sx, C]
            cur_size = (cur_size[0] // strides[0], cur_size[1] // strides[1])
            new_shape = (B, cur_size[0], strides[0], cur_size[1], strides[1], C)
            x = x.view(new_shape)

            # Move the patch stride into the batch dimension
            # For example in 2d: [B, Sy, Sx, H // Sy, W // Sx, C]
            L = len(new_shape)
            permute = [0] + list(range(2, L - 1, 2)) + list(range(1, L - 1, 2)) + [L - 1]
            x = x.permute(permute)

            # Now finally flatten the relevant dims into the batch dimension
            x = x.flatten(0, len(strides))
            B *= strides[0] * strides[1]

        x = x.reshape(-1, self.size[0] * self.size[1], C)

        return x


class Reroll(nn.Module):
    """
    Undo the "unroll" operation so that intermediate features are usable
    """

    def __init__(
        self,
        input_size: tuple[int, int],
        patch_stride: tuple[int, int],
        unroll_schedule: list[tuple[int, int]],
        stage_ends: list[int],
        q_pool: int,
    ):
        super().__init__()
        self.size = (input_size[0] // patch_stride[0], input_size[1] // patch_stride[1])

        # The first stage has to reverse everything
        # The next stage has to reverse all but the first unroll, etc.
        self.schedule = {}
        size = self.size
        for i in range(stage_ends[-1] + 1):
            self.schedule[i] = (unroll_schedule, size)
            # Schedule unchanged if no pooling at a stage end
            if i in stage_ends[:q_pool]:
                if len(unroll_schedule) > 0:
                    size = (size[0] // unroll_schedule[0][0], size[1] // unroll_schedule[0][1])

                unroll_schedule = unroll_schedule[1:]

    def forward(self, x: torch.Tensor, block_idx: int, mask: torch.Tensor = None) -> torch.Tensor:
        """
        Roll the given tensor back up to spatial order assuming it's from the given block.

        If no mask is provided:
            - Returns [B, H, W, C] for 2d, [B, T, H, W, C] for 3d, etc.
        If a mask is provided:
            - Returns [B, #MUs, MUy, MUx, C] for 2d, etc.
        """

        (schedule, size) = self.schedule[block_idx]
        (B, N, C) = x.size()

        D = len(size)
        cur_mu_shape = [1] * D

        for strides in schedule:
            # Extract the current patch from N
            x = x.view(B, *strides, N // math.prod(strides), *cur_mu_shape, C)

            # Move that patch into the current MU
            # Example in 2d: [B, Sy, Sx, N//(Sy*Sx), MUy, MUx, C] -> [B, N//(Sy*Sx), Sy, MUy, Sx, MUx, C]
            L = len(x.shape)
            permute = [0, 1 + D] + sum([list(p) for p in zip(range(1, 1 + D), range(1 + D + 1, L - 1))], []) + [L - 1]
            x = x.permute(permute)

            # Reshape to [B, N//(Sy*Sx), *MU, C]
            for i in range(D):
                cur_mu_shape[i] *= strides[i]

            x = x.reshape(B, -1, *cur_mu_shape, C)
            N = x.shape[1]

        # Current shape (e.g., 2d: [B, #MUy*#MUx, MUy, MUx, C])
        x = x.view(B, N, *cur_mu_shape, C)

        # If masked, return [B, #MUs, MUy, MUx, C]
        if mask is not None:
            return x

        # If not masked, we can return [B, H, W, C]
        x = undo_windowing(x, size, cur_mu_shape)

        return x


class MaskUnitAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        heads: int,
        q_stride: int,
        window_size: int,
        use_mask_unit_attn: bool,
    ) -> None:
        super().__init__()
        self.dim_out = dim_out
        self.heads = heads
        self.q_stride = q_stride
        self.head_dim = dim_out // heads
        self.scale = self.head_dim**-0.5
        self.window_size = window_size
        self.use_mask_unit_attn = use_mask_unit_attn

        self.qkv = nn.Linear(dim, 3 * dim_out)
        self.proj = nn.Linear(dim_out, dim_out)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, _) = x.shape
        if self.use_mask_unit_attn is True:
            num_windows = N // (self.q_stride * self.window_size)
        else:
            num_windows = 1

        qkv = self.qkv(x).reshape(B, -1, num_windows, 3, self.heads, self.head_dim).permute(3, 0, 4, 2, 1, 5)
        (q, k, v) = qkv.unbind(0)

        if self.q_stride > 1:
            # Refer to Unroll to see how this performs a maxpool-Nd
            q = q.view(B, self.heads, num_windows, self.q_stride, -1, self.head_dim).amax(dim=3)

        x = F.scaled_dot_product_attention(q, k, v, scale=self.scale)  # pylint: disable=not-callable

        x = x.transpose(1, 3).reshape(B, -1, self.dim_out)
        x = self.proj(x)

        return x


class HieraBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        heads: int,
        mlp_ratio: float,
        drop_path: float,
        q_stride: int,
        window_size: int,
        use_mask_unit_attn: bool,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out

        self.norm1 = nn.LayerNorm(dim)
        if dim != dim_out:
            self.proj = nn.Linear(dim, dim_out)
        else:
            self.proj = None

        self.attn = MaskUnitAttention(dim, dim_out, heads, q_stride, window_size, use_mask_unit_attn)
        self.norm2 = nn.LayerNorm(dim_out)
        self.mlp = MLP(dim_out, [int(dim_out * mlp_ratio), dim_out], activation_layer=nn.GELU)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_norm = self.norm1(x)
        if self.proj is not None:
            x = self.proj(x_norm)
            x = x.view(x.shape[0], self.attn.q_stride, -1, x.shape[-1]).amax(dim=1)  # Max-pool

        x = x + self.drop_path(self.attn(x_norm))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


# pylint: disable=too-many-instance-attributes
class Hiera(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin):
    scriptable = False
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        mask_unit_size = (8, 8)
        q_stride = (2, 2)
        q_pool = 3
        patch_kernel = (7, 7)
        patch_stride = (4, 4)
        patch_padding = (3, 3)
        mask_unit_attn = (True, True, False, False)
        dim_mul = 2.0
        head_mul = 2.0
        mlp_ratio = 4.0
        depths: list[int] = self.config["depths"]
        embed_dim: int = self.config["embed_dim"]
        num_heads: int = self.config["num_heads"]
        abs_win_pos_embed: bool = self.config["abs_win_pos_embed"]
        drop_path_rate: float = self.config["drop_path_rate"]

        tokens_spatial_shape = [i // s for i, s in zip(image_size, patch_stride)]
        num_tokens = math.prod(tokens_spatial_shape)
        flat_mu_size = math.prod(mask_unit_size)
        flat_q_stride = math.prod(q_stride)
        assert q_pool < len(depths)
        assert patch_stride[0] == patch_stride[1]

        self.tokens_spatial_shape = tokens_spatial_shape
        self.mask_unit_size = mask_unit_size
        self.patch_stride = patch_stride
        self.q_stride = q_stride
        self.q_pool = q_pool
        self.mu_size = flat_mu_size
        self.mask_spatial_shape = [i // s for i, s in zip(tokens_spatial_shape, mask_unit_size)]
        self.stage_ends = [sum(depths[:i]) - 1 for i in range(1, len(depths) + 1)]
        self.num_special_tokens = 0
        self.num_layers = sum(depths)

        stem_dim = embed_dim
        self.stem = PatchEmbed(self.input_channels, stem_dim, patch_kernel, patch_stride, patch_padding)

        global_pos_size = (image_size[0] // 2**4, image_size[1] // 2**4)
        if abs_win_pos_embed is True:
            # AbsWin variant
            self.pos_embed = nn.Parameter(torch.zeros(1, embed_dim, *global_pos_size))
            self.pos_embed_win = nn.Parameter(torch.zeros(1, embed_dim, *mask_unit_size))
        else:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_tokens, embed_dim))
            self.pos_embed_win = None

        self.unroll = Unroll(image_size, patch_stride, [q_stride] * len(self.stage_ends[:-1]))
        self.reroll = Reroll(image_size, patch_stride, [q_stride] * len(self.stage_ends[:-1]), self.stage_ends, q_pool)

        q_pool_blocks = [x + 1 for x in self.stage_ends[:q_pool]]
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, self.num_layers)]  # Stochastic depth decay rule

        cur_stage = 0
        layers = []
        self.block_count = []
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(self.num_layers):
            dim_out = embed_dim

            # Mask unit or global attention.
            # Lag by 1 block, so that global attention, applied post pooling on lower resolution
            use_mask_unit_attn = mask_unit_attn[cur_stage]

            if i - 1 in self.stage_ends:
                dim_out = int(embed_dim * dim_mul)
                num_heads = int(num_heads * head_mul)
                cur_stage += 1
                if i in q_pool_blocks:
                    flat_mu_size //= flat_q_stride

            layers.append(
                HieraBlock(
                    dim=embed_dim,
                    dim_out=dim_out,
                    heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    drop_path=dpr[i],
                    q_stride=(flat_q_stride if i in q_pool_blocks else 1),
                    window_size=flat_mu_size,
                    use_mask_unit_attn=use_mask_unit_attn,
                )
            )
            embed_dim = dim_out

            if i in self.stage_ends:
                stages[f"stage{cur_stage+1}"] = nn.Sequential(*layers)
                return_channels.append(dim_out)
                self.block_count.append(i)
                layers = []

        assert len(layers) == 0

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.LayerNorm(embed_dim),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_stride[0] * 8
        self.stem_stride = patch_stride[0]
        self.stem_width = stem_dim
        self.encoding_size = embed_dim
        # self.decoder_block

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.02)

            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0.02)
                nn.init.ones_(m.weight)

        nn.init.trunc_normal_(self.pos_embed, std=0.02)
        if self.pos_embed_win is not None:
            nn.init.trunc_normal_(self.pos_embed_win, std=0.02)
        if isinstance(self.classifier, nn.Linear):
            self.classifier.weight.data.mul_(0.001)
            self.classifier.bias.data.mul_(0.001)

    def _get_pos_embed(self) -> torch.Tensor:
        if self.pos_embed_win is not None:
            # Reference interpolation adapted from:
            # https://github.com/facebookresearch/sam2/blob/main/sam2/modeling/backbones/hieradet.py#L273
            pos_embed_win = self.pos_embed_win.tile(self.mask_spatial_shape)
            pos_embed = F.interpolate(
                self.pos_embed,
                size=pos_embed_win.shape[-2:],
                mode="bicubic",
                antialias=True,
            )
            pos_embed = pos_embed + pos_embed_win
            pos_embed = pos_embed.flatten(2).transpose(1, 2)
        else:
            pos_embed = self.pos_embed

        return pos_embed

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)
        x = x + self._get_pos_embed()
        x = self.unroll(x)

        out = {}
        for idx, (name, module) in enumerate(self.body.named_children()):
            x = module(x)
            if name in self.return_stages:
                z = self.reroll(x, self.block_count[idx])
                out[name] = z.permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        torch._assert(return_all_features is False, "not supported")  # pylint: disable=protected-access
        if ids_keep is not None:
            num_windows = math.prod(self.mask_spatial_shape)

            # Mask already in opposite form
            mask = mask_from_indices(ids_keep, num_windows)
            mask = mask.bool()
            patch_mask = mask.view(x.shape[0], 1, *self.mask_spatial_shape)  # B, C, *mask_spatial_shape
        else:
            patch_mask = None

        x = self.stem(x, patch_mask)
        x = x + self._get_pos_embed()
        x = self.unroll(x)

        # Discard masked tokens
        if ids_keep is not None:
            x = x[mask[..., None].tile(1, self.mu_size, x.shape[2])].view(x.shape[0], -1, x.shape[-1])

        x = self.body(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            x = x.mean(dim=1)
            x = self.features(x)
            result["embedding"] = x

        return result

    def masked_encoding(self, x: torch.Tensor, mask: torch.Tensor) -> tuple[list[torch.Tensor], torch.Tensor]:
        # Binary mask: 1 is *keep*, 0 is *remove*
        # Note this is opposite to original MAE
        mask = 1 - mask

        # Un-shuffle to get the binary mask
        mask = mask.bool()
        patch_mask = mask.view(x.shape[0], 1, *self.mask_spatial_shape)  # B, C, *mask_spatial_shape

        x = self.stem(x, patch_mask)
        x = x + self._get_pos_embed()
        x = self.unroll(x)

        # Discard masked tokens
        x = x[mask[..., None].tile(1, self.mu_size, x.shape[2])].view(x.shape[0], -1, x.shape[-1])

        out = []
        for idx, (name, module) in enumerate(self.body.named_children()):
            x = module(x)
            if name in self.return_stages:
                out.append(self.reroll(x, self.block_count[idx], mask=mask))

        return (out, mask)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        x = x + self._get_pos_embed()
        x = self.unroll(x)
        x = self.body(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = x.mean(dim=1)
        x = self.features(x)

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        if self.pos_embed_win is not None:
            global_pos_size = (new_size[0] // 2**4, new_size[1] // 2**4)
            pos_embed = F.interpolate(
                self.pos_embed,
                size=global_pos_size,
                mode="bicubic",
                antialias=True,
            )
            self.pos_embed = nn.Parameter(pos_embed)

        else:
            self.pos_embed = nn.Parameter(
                adjust_position_embedding(
                    self.pos_embed,
                    (old_size[0] // self.patch_stride[0], old_size[1] // self.patch_stride[1]),
                    (new_size[0] // self.patch_stride[0], new_size[1] // self.patch_stride[1]),
                    0,
                )
            )

        # Re-init vars
        self.tokens_spatial_shape = [i // s for i, s in zip(new_size, self.patch_stride)]
        self.mask_spatial_shape = [i // s for i, s in zip(self.tokens_spatial_shape, self.mask_unit_size)]

        # Re-init rolls
        self.unroll = Unroll(new_size, self.patch_stride, [self.q_stride] * len(self.stage_ends[:-1]))
        self.reroll = Reroll(
            new_size,
            self.patch_stride,
            [self.q_stride] * len(self.stage_ends[:-1]),
            self.stage_ends,
            self.q_pool,
        )


registry.register_model_config(
    "hiera_tiny",
    Hiera,
    config={
        "depths": [1, 2, 7, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_small",
    Hiera,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_base",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_base_plus",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 112,
        "num_heads": 2,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_large",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "num_heads": 2,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "hiera_huge",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 256,
        "num_heads": 4,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.3,
    },
)

# AbsWin variant
registry.register_model_config(
    "hiera_abswin_tiny",
    Hiera,
    config={
        "depths": [1, 2, 7, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_small",
    Hiera,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_base",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_base_plus",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 112,
        "num_heads": 2,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_large",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "num_heads": 2,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "hiera_abswin_huge",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 256,
        "num_heads": 4,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.3,
    },
)

registry.register_weights(
    "hiera_abswin_base_mim",
    {
        "url": "https://huggingface.co/birder-project/hiera_abswin_base_mim/resolve/main",
        "description": (
            "Hiera base with abswin image encoder pre-trained using Masked Image Modeling (MIM) for 400 epochs. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 192.7,
                "sha256": "d8ce30c2d9d46af8bdfe7c591610d12c4e592ad4bc4b00e2ba41457e683702da",
            },
            "safetensors": {
                "file_size": 192.6,
                "sha256": "b03f9ec88f7b391fe6b2f5ab7707c156351ab4151c95163f631db0712c48b843",
            },
        },
        "net": {"network": "hiera_abswin_base", "tag": "mim"},
    },
)
registry.register_weights(
    "hiera_abswin_base_mim-intermediate-eu-common",
    {
        "url": "https://huggingface.co/birder-project/hiera_abswin_base_mim-intermediate-eu-common/resolve/main",
        "description": (
            "Hiera base with abswin model with MIM pretraining and intermediate training, "
            "then fine-tuned on the eu-common dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 194.9,
                "sha256": "e6e9d28e0cd98a5c72c1a4a67301c62d2ade97e6d162f34ad630e5ffb2f9ec2a",
            }
        },
        "net": {"network": "hiera_abswin_base", "tag": "mim-intermediate-eu-common"},
    },
)



import logging
import unittest

import numpy as np
import torch
from torch import nn

from birder import net
from birder.inference import classification
from birder.inference.data_parallel import InferenceDataParallel

logging.disable(logging.CRITICAL)


class OrderTestModel(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Return first element of each sample as identifier
        # This lets us track if order is preserved
        batch_size = x.size(0)
        return x.view(batch_size, -1)[:, :10]


class TestInference(unittest.TestCase):
    def setUp(self) -> None:
        self.size = net.GhostNet_v2.default_size
        self.num_classes = 10
        self.model = net.GhostNet_v2(3, self.num_classes, config={"width": 1.0})
        self.model.eval()

    def test_infer_batch_default_behavior(self) -> None:
        with torch.inference_mode():
            (out, embed) = classification.infer_batch(self.model, torch.rand((1, 3, *self.size)))

        self.assertIsNone(embed)
        self.assertEqual(len(out), 1)
        self.assertEqual(len(out[0]), self.num_classes)
        self.assertAlmostEqual(out[0].sum(), 1.0, places=5)

    def test_infer_batch_return_embedding(self) -> None:
        with torch.inference_mode():
            (out, embed) = classification.infer_batch(self.model, torch.rand((1, 3, *self.size)), return_embedding=True)

        self.assertIsNotNone(embed)
        self.assertEqual(embed.shape[0], 1)  # type: ignore[union-attr]
        self.assertEqual(embed.shape[1], self.model.embedding_size)  # type: ignore[union-attr]
        self.assertEqual(len(out), 1)
        self.assertEqual(len(out[0]), self.num_classes)
        self.assertAlmostEqual(out[0].sum(), 1.0, places=5)

    def test_infer_batch_tta(self) -> None:
        with torch.inference_mode():
            (out, embed) = classification.infer_batch(self.model, torch.rand((1, 3, *self.size)), tta=True)

        self.assertIsNone(embed)
        self.assertEqual(len(out), 1)
        self.assertEqual(len(out[0]), self.num_classes)
        self.assertAlmostEqual(out[0].sum(), 1.0, places=5)

    def test_infer_batch_return_logits(self) -> None:
        dummy_input = torch.rand((1, 3, *self.size))
        with torch.inference_mode():
            (out, embed) = classification.infer_batch(self.model, dummy_input, return_logits=True)

        self.assertIsNone(embed)
        self.assertEqual(len(out), 1)
        self.assertEqual(len(out[0]), self.num_classes)

        # Logits should NOT sum to 1
        self.assertNotAlmostEqual(out[0].sum(), 1.0, places=5)

        with torch.inference_mode():
            # Verify that the output is indeed logits by comparing to a manual forward pass
            expected_logits = self.model(dummy_input).cpu().float().numpy()

        np.testing.assert_allclose(out, expected_logits)


@unittest.skipUnless(torch.cuda.is_available(), "CUDA not available")
class TestInferenceDataParallel(unittest.TestCase):
    def setUp(self) -> None:
        self.size = net.GhostNet_v2.default_size
        self.num_classes = 10
        self.device = torch.device("cuda")
        self.num_devices = torch.cuda.device_count()

        # Create and prepare model for inference (mimics load_model with inference=True)
        self.model = net.GhostNet_v2(3, self.num_classes, config={"width": 1.0})
        self.model.eval()
        for param in self.model.parameters():
            param.requires_grad = False

    @unittest.skipUnless(torch.cuda.device_count() >= 2, "Requires at least 2 GPUs")
    def test_basic_forward(self) -> None:
        model_parallel = InferenceDataParallel(self.model)

        # Test different batch sizes
        batch_sizes = [1, 4, 7, 16, 31]  # Include prime numbers and edge cases

        for batch_size in batch_sizes:
            with self.subTest(batch_size=batch_size):
                x = torch.randn(batch_size, 3, *self.size)
                with torch.inference_mode():
                    # Single GPU reference
                    model_single = self.model.to(self.device)
                    out_single = model_single(x.to(self.device))

                    # Multi-GPU
                    out_parallel = model_parallel(x)

                # Check shapes match
                self.assertEqual(out_single.size(), out_parallel.size())

                # Check outputs are close
                diff = (out_single - out_parallel).abs().max()
                self.assertLess(diff.item(), 1e-5, f"Output mismatch for batch_size={batch_size}")

        # Test with CPU output
        model_parallel = InferenceDataParallel(self.model, output_device="cpu")
        x = torch.randn(batch_size, 3, *self.size)
        with torch.inference_mode():
            out_parallel = model_parallel(x)

        self.assertEqual(out_parallel.device.type, "cpu")

    @unittest.skipUnless(torch.cuda.device_count() >= 2, "Requires at least 2 GPUs")
    def test_output_order_preservation(self) -> None:
        model = OrderTestModel()
        model.eval()
        for param in model.parameters():
            param.requires_grad = False

        model_parallel = InferenceDataParallel(model)

        # Create input where each sample is identifiable by its values
        batch_size = 17
        x = torch.zeros(batch_size, 3, 32, 32)
        for i in range(batch_size):
            x[i, :, :, :] = float(i)  # Each sample has unique value

        with torch.inference_mode():
            # Single GPU reference
            model_single = model.to(self.device)
            out_single = model_single(x.to(self.device))

            # Multi-GPU
            out_parallel = model_parallel(x)

        # Verify exact order preservation
        for i in range(batch_size):
            self.assertAlmostEqual(out_parallel[i, 0].item(), float(i), places=6)

        # Also verify against single GPU output
        np.testing.assert_allclose(out_single.cpu().numpy(), out_parallel.cpu().numpy(), rtol=1e-6)

    @unittest.skipUnless(torch.cuda.device_count() >= 2, "Requires at least 2 GPUs")
    def test_custom_func(self) -> None:
        model_parallel = InferenceDataParallel(self.model)

        batch_size = 8
        x = torch.randn(batch_size, 3, *self.size)
        with torch.inference_mode():
            embeddings = model_parallel.embedding(x)
            self.assertEqual(embeddings.size(0), batch_size)
            self.assertEqual(embeddings.size(1), self.model.embedding_size)

            logits = model_parallel.classify(embeddings)
            self.assertEqual(logits.size(), (batch_size, self.num_classes))

    @unittest.skipUnless(torch.cuda.device_count() >= 2, "Requires at least 2 GPUs")
    def test_integration_with_infer_batch(self) -> None:
        model_parallel = InferenceDataParallel(self.model)

        batch_size = 4
        x = torch.randn(batch_size, 3, *self.size)
        with torch.inference_mode():
            (out, embed) = classification.infer_batch(model_parallel, x)
            self.assertIsNone(embed)
            self.assertEqual(out.shape, (batch_size, self.num_classes))

            # Test with embeddings
            (out, embed) = classification.infer_batch(model_parallel, x, return_embedding=True)
            self.assertIsNotNone(embed)
            self.assertEqual(embed.shape, (batch_size, self.model.embedding_size))  # type: ignore[union-attr]

            # Test with TTA
            (out, embed) = classification.infer_batch(model_parallel, x, tta=True)
            self.assertEqual(out.shape, (batch_size, self.num_classes))

            # Test with logits
            (out, embed) = classification.infer_batch(model_parallel, x, return_logits=True)
            self.assertEqual(out.shape, (batch_size, self.num_classes))

    def test_single_gpu_fallback(self) -> None:
        model_parallel = InferenceDataParallel(self.model, device_ids=[0])
        batch_size = 4
        x = torch.randn(batch_size, 3, *self.size)
        with torch.inference_mode():
            out = model_parallel(x)

        self.assertEqual(out.size(), (batch_size, self.num_classes))

        # Verify it produces same output as direct model
        with torch.inference_mode():
            model_single = self.model.to(self.device)
            out_direct = model_single(x.to(self.device))

        np.testing.assert_allclose(out.cpu().numpy(), out_direct.cpu().numpy(), rtol=1e-5)



"""
ConvNeXt v2, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/convnext.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/convnext.py

Paper "ConvNeXt V2: Co-designing and Scaling ConvNets with Masked Autoencoders",
https://arxiv.org/abs/2301.00808
"""

# Reference license: BSD 3-Clause and Apache-2.0

from collections import OrderedDict
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


class GRN(nn.Module):
    """
    GRN (Global Response Normalization) layer
    """

    def __init__(self, dim: int):
        super().__init__()
        self.gamma = nn.Parameter(torch.zeros(1, 1, 1, dim))
        self.beta = nn.Parameter(torch.zeros(1, 1, 1, dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        gx = torch.linalg.vector_norm(x, ord=2, dim=(1, 2), keepdim=True)  # pylint: disable=not-callable
        nx = gx / (gx.mean(dim=-1, keepdim=True) + 1e-6)

        return self.gamma * (x * nx) + self.beta + x


class ConvNeXtBlock(nn.Module):
    def __init__(
        self,
        channels: int,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()
        self.block = nn.Sequential(
            nn.Conv2d(
                channels,
                channels,
                kernel_size=(7, 7),
                stride=(1, 1),
                padding=(3, 3),
                groups=channels,
                bias=True,
            ),
            Permute([0, 2, 3, 1]),
            nn.LayerNorm(channels, eps=1e-6),
            nn.Linear(channels, 4 * channels),  # Same as 1x1 conv
            nn.GELU(),
            GRN(4 * channels),
            nn.Linear(4 * channels, channels),  # Same as 1x1 conv
            Permute([0, 3, 1, 2]),
        )
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.block(x)
        x = self.stochastic_depth(x)
        x += identity

        return x


# pylint: disable=invalid-name
class ConvNeXt_v2(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        in_channels: list[int] = self.config["in_channels"]
        num_layers: list[int] = self.config["num_layers"]
        drop_path_rate: float = self.config["drop_path_rate"]
        out_channels = in_channels[1:] + [-1]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            in_channels[0],
            kernel_size=(4, 4),
            stride=(4, 4),
            padding=(0, 0),
            bias=True,
            norm_layer=LayerNorm2d,
            activation_layer=None,
        )

        total_stage_blocks = sum(num_layers)
        stage_block_id = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        layers = []
        for idx, (i, out, n) in enumerate(zip(in_channels, out_channels, num_layers)):
            # Bottlenecks
            for _ in range(n):
                # Adjust stochastic depth probability based on the depth of the stage block
                sd_prob = drop_path_rate * stage_block_id / (total_stage_blocks - 1.0)
                layers.append(ConvNeXtBlock(i, sd_prob))
                stage_block_id += 1

            stages[f"stage{idx+1}"] = nn.Sequential(*layers)
            return_channels.append(i)
            layers = []

            # Down sampling
            if out != -1:
                layers.append(
                    nn.Sequential(
                        LayerNorm2d(i, eps=1e-6),
                        nn.Conv2d(i, out, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0), bias=True),
                    )
                )

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(in_channels[-1], eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = in_channels[-1]
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = in_channels[0]
        self.encoding_size = in_channels[-1]
        self.decoder_block = partial(ConvNeXtBlock, stochastic_depth_prob=0)

        # Weights initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

        if isinstance(self.classifier, nn.Linear):
            self.classifier.weight.data.mul_(0.001)
            self.classifier.bias.data.mul_(0.001)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "convnext_v2_atto",
    ConvNeXt_v2,
    config={"in_channels": [40, 80, 160, 320], "num_layers": [2, 2, 6, 2], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "convnext_v2_femto",
    ConvNeXt_v2,
    config={"in_channels": [48, 96, 192, 384], "num_layers": [2, 2, 6, 2], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "convnext_v2_pico",
    ConvNeXt_v2,
    config={"in_channels": [64, 128, 256, 512], "num_layers": [2, 2, 6, 2], "drop_path_rate": 0.0},
)
registry.register_model_config(
    "convnext_v2_nano",
    ConvNeXt_v2,
    config={"in_channels": [80, 160, 320, 640], "num_layers": [2, 2, 8, 2], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "convnext_v2_tiny",
    ConvNeXt_v2,
    config={"in_channels": [96, 192, 384, 768], "num_layers": [3, 3, 9, 3], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "convnext_v2_small",  # Not in the original v2, taken from v1
    ConvNeXt_v2,
    config={"in_channels": [96, 192, 384, 768], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "convnext_v2_base",
    ConvNeXt_v2,
    config={"in_channels": [128, 256, 512, 1024], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "convnext_v2_large",
    ConvNeXt_v2,
    config={"in_channels": [192, 384, 768, 1536], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "convnext_v2_huge",
    ConvNeXt_v2,
    config={"in_channels": [352, 704, 1408, 2816], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.3},
)

registry.register_weights(
    "convnext_v2_atto_il-common",
    {
        "description": "ConvNeXt v2 nano model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 13.4,
                "sha256": "b001b4d71c74966b530f847490edf1c4c7d769a34fa29aa96238602e3248a2a2",
            }
        },
        "net": {"network": "convnext_v2_atto", "tag": "il-common"},
    },
)
registry.register_weights(
    "convnext_v2_tiny_intermediate-il-common",
    {
        "url": "https://huggingface.co/birder-project/convnext_v2_tiny_intermediate-il-common/resolve/main",
        "description": "ConvNeXt v2 tiny model with intermediate training, then fine-tuned on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 107.5,
                "sha256": "590b74dbf39aa1d6356b9001d5505c00ac97b7399fa994efc58c7e14fd4f668a",
            }
        },
        "net": {"network": "convnext_v2_tiny", "tag": "intermediate-il-common"},
    },
)
registry.register_weights(
    "convnext_v2_tiny_intermediate-eu-common",
    {
        "url": "https://huggingface.co/birder-project/convnext_v2_tiny_intermediate-eu-common/resolve/main",
        "description": "ConvNeXt v2 tiny model with intermediate training, then fine-tuned on the eu-common dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 108.5,
                "sha256": "a0da4037acda6c50c9032a6462a810082aa79193bdec26203d98fdfdad699fc4",
            }
        },
        "net": {"network": "convnext_v2_tiny", "tag": "intermediate-eu-common"},
    },
)



"""
RegNet, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/regnet.py

Paper "Designing Network Design Spaces", https://arxiv.org/abs/2003.13678
"""

# Reference license: BSD 3-Clause

import math
from collections import OrderedDict
from collections.abc import Callable
from collections.abc import Iterator
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.base import make_divisible


class BlockParams:
    def __init__(
        self,
        depths: list[int],
        widths: list[int],
        group_widths: list[int],
        bottleneck_multipliers: list[float],
        strides: list[int],
        se_ratio: float,
    ) -> None:
        self.depths = depths
        self.widths = widths
        self.group_widths = group_widths
        self.bottleneck_multipliers = bottleneck_multipliers
        self.strides = strides
        self.se_ratio = se_ratio

    @classmethod
    def from_init_params(
        cls,
        depth: int,
        w_0: int,
        w_a: float,
        w_m: float,
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: float,
    ) -> "BlockParams":
        """
        Programmatically compute all the per-block settings, given the RegNet parameters.

        The first step is to compute the quantized linear block parameters,
        in log space. Key parameters are:
        - w_a is the width progression slope
        - w_0 is the initial width
        - w_m is the width stepping in the log space

        In other terms
        log(block_width) = log(w_0) + w_m * block_capacity,
        with bock_capacity ramping up following the w_0 and w_a params.
        This block width is finally quantized to multiples of 8.
        The second step is to compute the parameters per stage,
        taking into account the skip connection and the final 1x1 convolutions.
        We use the fact that the output width is constant within a stage.
        """

        QUANT = 8  # pylint: disable=invalid-name
        STRIDE = 2  # pylint: disable=invalid-name

        if w_a < 0 or w_0 <= 0 or w_m <= 1 or w_0 % 8 != 0:
            raise ValueError("Invalid RegNet settings")

        # Compute the block widths. Each stage has one unique block width
        widths_cont = torch.arange(depth) * w_a + w_0
        block_capacity = torch.round(torch.log(widths_cont / w_0) / math.log(w_m))
        block_widths = (torch.round(torch.divide(w_0 * torch.pow(w_m, block_capacity), QUANT)) * QUANT).int().tolist()
        num_stages = len(set(block_widths))

        # Convert to per stage parameters
        split_helper = zip(block_widths + [0], [0] + block_widths, block_widths + [0], [0] + block_widths)
        splits = [w != wp or r != rp for w, wp, r, rp in split_helper]

        stage_widths = [w for w, t in zip(block_widths, splits[:-1]) if t]
        stage_depths = torch.diff(torch.tensor([d for d, t in enumerate(splits) if t])).int().tolist()

        strides = [STRIDE] * num_stages
        bottleneck_multipliers = [bottleneck_multiplier] * num_stages
        group_widths = [group_width] * num_stages

        # Adjust the compatibility of stage widths and group widths
        (stage_widths, group_widths) = cls._adjust_widths_groups_compatibility(
            stage_widths, bottleneck_multipliers, group_widths
        )

        return cls(
            depths=stage_depths,
            widths=stage_widths,
            group_widths=group_widths,
            bottleneck_multipliers=bottleneck_multipliers,
            strides=strides,
            se_ratio=se_ratio,
        )

    def _get_expanded_params(self) -> Iterator[tuple[int, int, int, int, float]]:
        return zip(self.widths, self.strides, self.depths, self.group_widths, self.bottleneck_multipliers)

    @staticmethod
    def _adjust_widths_groups_compatibility(
        stage_widths: list[int], bottleneck_ratios: list[float], group_widths: list[int]
    ) -> tuple[list[int], list[int]]:
        """
        Adjusts the compatibility of widths and groups,
        depending on the bottleneck ratio.
        """

        # Compute all widths for the current settings
        widths = [int(w * b) for w, b in zip(stage_widths, bottleneck_ratios)]
        group_widths_min = [min(g, w_bot) for g, w_bot in zip(group_widths, widths)]

        # Compute the adjusted widths so that stage and group widths fit
        ws_bot = [make_divisible(w_bot, g) for w_bot, g in zip(widths, group_widths_min)]
        stage_widths = [int(w_bot / b) for w_bot, b in zip(ws_bot, bottleneck_ratios)]
        return stage_widths, group_widths_min


class BottleneckTransform(nn.Module):
    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: tuple[int, int],
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: float,
        activation_layer: Callable[..., nn.Module] = nn.ReLU,
    ) -> None:
        super().__init__()
        w_b = int(round(width_out * bottleneck_multiplier))
        g = w_b // group_width
        width_se_out = int(round(se_ratio * width_in))

        if width_se_out == 0:
            se_block = nn.Identity()
        else:
            se_block = SqueezeExcitation(input_channels=w_b, squeeze_channels=width_se_out, activation=activation_layer)

        self.block = nn.Sequential(
            Conv2dNormActivation(
                width_in,
                w_b,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                activation_layer=activation_layer,
                bias=False,
            ),
            Conv2dNormActivation(
                w_b,
                w_b,
                kernel_size=(3, 3),
                stride=stride,
                padding=(1, 1),
                groups=g,
                activation_layer=activation_layer,
                bias=False,
            ),
            se_block,
            Conv2dNormActivation(
                w_b, width_out, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False, activation_layer=None
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.block(x)


class ResBottleneckBlock(nn.Module):
    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: tuple[int, int],
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: float,
    ) -> None:
        super().__init__()

        if width_in != width_out or stride[0] != 1 or stride[1] != 1:
            self.proj = Conv2dNormActivation(
                width_in,
                width_out,
                kernel_size=(1, 1),
                stride=stride,
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            )

        else:
            self.proj = nn.Identity()

        self.f = BottleneckTransform(
            width_in,
            width_out,
            stride=stride,
            group_width=group_width,
            bottleneck_multiplier=bottleneck_multiplier,
            se_ratio=se_ratio,
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x) + self.f(x)
        x = self.relu(x)

        return x


class AnyStage(nn.Module):
    def __init__(
        self,
        width_in: int,
        width_out: int,
        stride: tuple[int, int],
        depth: int,
        group_width: int,
        bottleneck_multiplier: float,
        se_ratio: float,
    ) -> None:
        super().__init__()

        layers = []
        for i in range(depth):
            if i == 0:
                in_ch = width_in
                cur_stride = stride
            else:
                in_ch = width_out
                cur_stride = (1, 1)

            layers.append(
                ResBottleneckBlock(
                    in_ch,
                    width_out,
                    stride=cur_stride,
                    group_width=group_width,
                    bottleneck_multiplier=bottleneck_multiplier,
                    se_ratio=se_ratio,
                )
            )

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.block(x)


class RegNet(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.block\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        stem_width = 32
        bottleneck_multiplier = 1.0
        depth: int = self.config["depth"]
        w_0: int = self.config["w_0"]
        w_a: float = self.config["w_a"]
        w_m: float = self.config["w_m"]
        group_width: int = self.config["group_width"]
        se_ratio: float = self.config["se_ratio"]

        block_params = BlockParams.from_init_params(depth, w_0, w_a, w_m, group_width, bottleneck_multiplier, se_ratio)

        self.stem = Conv2dNormActivation(
            self.input_channels, stem_width, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False
        )

        current_width = stem_width
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        i = 0
        for (
            width_out,
            stride,
            depth,
            group_width,
            bottleneck_multiplier,
        ) in block_params._get_expanded_params():
            stages[f"stage{i+1}"] = AnyStage(
                current_width, width_out, (stride, stride), depth, group_width, bottleneck_multiplier, se_ratio
            )
            return_channels.append(width_out)
            current_width = width_out
            i += 1

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = current_width
        self.classifier = self.create_classifier()

        self.stem_stride = 2
        self.stem_width = stem_width
        self.encoding_size = current_width
        decoder_block = partial(
            BottleneckTransform,
            stride=(1, 1),
            group_width=64,
            bottleneck_multiplier=1.0,
            se_ratio=se_ratio,
        )
        self.decoder_block = lambda x: decoder_block(x, x)

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # Note that there is no bias due to BN
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                nn.init.normal_(m.weight, mean=0.0, std=math.sqrt(2.0 / fan_out))

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, mean=0.0, std=0.01)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


# RegNet X
registry.register_model_config(
    "regnet_x_200m",
    RegNet,
    config={"depth": 13, "w_0": 24, "w_a": 36.44, "w_m": 2.49, "group_width": 8, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_400m",
    RegNet,
    config={"depth": 22, "w_0": 24, "w_a": 24.48, "w_m": 2.54, "group_width": 16, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_600m",
    RegNet,
    config={"depth": 16, "w_0": 48, "w_a": 36.97, "w_m": 2.24, "group_width": 24, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_800m",
    RegNet,
    config={"depth": 16, "w_0": 56, "w_a": 35.73, "w_m": 2.28, "group_width": 16, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_1_6g",
    RegNet,
    config={"depth": 18, "w_0": 80, "w_a": 34.01, "w_m": 2.25, "group_width": 24, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_3_2g",
    RegNet,
    config={"depth": 25, "w_0": 88, "w_a": 26.31, "w_m": 2.25, "group_width": 48, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_4g",
    RegNet,
    config={"depth": 23, "w_0": 96, "w_a": 38.65, "w_m": 2.43, "group_width": 40, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_6_4g",
    RegNet,
    config={"depth": 17, "w_0": 184, "w_a": 60.83, "w_m": 2.07, "group_width": 56, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_8g",
    RegNet,
    config={"depth": 23, "w_0": 80, "w_a": 49.56, "w_m": 2.88, "group_width": 120, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_12g",
    RegNet,
    config={"depth": 19, "w_0": 168, "w_a": 73.36, "w_m": 2.37, "group_width": 112, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_16g",
    RegNet,
    config={"depth": 22, "w_0": 216, "w_a": 55.59, "w_m": 2.1, "group_width": 128, "se_ratio": 0.0},
)
registry.register_model_config(
    "regnet_x_32g",
    RegNet,
    config={"depth": 23, "w_0": 320, "w_a": 69.86, "w_m": 2.0, "group_width": 168, "se_ratio": 0.0},
)

# RegNet Y
registry.register_model_config(
    "regnet_y_200m",
    RegNet,
    config={"depth": 13, "w_0": 24, "w_a": 36.44, "w_m": 2.49, "group_width": 8, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_400m",
    RegNet,
    config={"depth": 16, "w_0": 48, "w_a": 27.89, "w_m": 2.09, "group_width": 8, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_600m",
    RegNet,
    config={"depth": 15, "w_0": 48, "w_a": 32.54, "w_m": 2.32, "group_width": 16, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_800m",
    RegNet,
    config={"depth": 14, "w_0": 56, "w_a": 38.84, "w_m": 2.4, "group_width": 16, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_1_6g",
    RegNet,
    config={"depth": 27, "w_0": 48, "w_a": 20.71, "w_m": 2.65, "group_width": 24, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_3_2g",
    RegNet,
    config={"depth": 21, "w_0": 80, "w_a": 42.63, "w_m": 2.66, "group_width": 24, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_4g",
    RegNet,
    config={"depth": 22, "w_0": 96, "w_a": 31.41, "w_m": 2.24, "group_width": 64, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_6_4g",
    RegNet,
    config={"depth": 25, "w_0": 112, "w_a": 33.22, "w_m": 2.27, "group_width": 72, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_8g",
    RegNet,
    config={"depth": 17, "w_0": 192, "w_a": 76.82, "w_m": 2.19, "group_width": 56, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_12g",
    RegNet,
    config={"depth": 19, "w_0": 168, "w_a": 73.36, "w_m": 2.37, "group_width": 112, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_16g",
    RegNet,
    config={"depth": 18, "w_0": 200, "w_a": 106.23, "w_m": 2.48, "group_width": 112, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_32g",
    RegNet,
    config={"depth": 20, "w_0": 232, "w_a": 115.89, "w_m": 2.53, "group_width": 232, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_64g",
    RegNet,
    config={"depth": 20, "w_0": 352, "w_a": 147.48, "w_m": 2.4, "group_width": 328, "se_ratio": 0.25},
)
registry.register_model_config(
    "regnet_y_128g",
    RegNet,
    config={"depth": 27, "w_0": 456, "w_a": 160.83, "w_m": 2.52, "group_width": 264, "se_ratio": 0.25},
)

registry.register_weights(
    "regnet_x_400m_il-common",
    {
        "description": "RegNet X 400m model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 20.3,
                "sha256": "c20ea88412fd2d85e28426dcb8300d38bc0fea856e7d6afff2a616fa0363c06d",
            }
        },
        "net": {"network": "regnet_x_400m", "tag": "il-common"},
    },
)

registry.register_weights(
    "regnet_y_200m_il-common",
    {
        "description": "RegNet Y 200m model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.4,
                "sha256": "0761351fc6624e476b59828efb88859f3667943aca918f732817b2d35a108884",
            }
        },
        "net": {"network": "regnet_y_200m", "tag": "il-common"},
    },
)
registry.register_weights(
    "regnet_y_400m_il-common",
    {
        "description": "RegNet Y 400m model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 15.8,
                "sha256": "20cd73f54aff3fddff410cba7686da8e4768f74c7a4981cfde4217e7e58eb71b",
            }
        },
        "net": {"network": "regnet_y_400m", "tag": "il-common"},
    },
)
registry.register_weights(
    "regnet_y_600m_il-common",
    {
        "description": "RegNet Y 600m model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 21.9,
                "sha256": "7ff9dc09d0f3216d85fe739fd0977b98a96e394d9a8916afa5454aa8fedda4c6",
            }
        },
        "net": {"network": "regnet_y_600m", "tag": "il-common"},
    },
)
registry.register_weights(
    "regnet_y_1_6g_il-common",
    {
        "description": "RegNet Y 1.6g model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 41.0,
                "sha256": "c2f8b17559a7d55efd1119352612a93063b393e0f05e8eb7ab0b5f637344cacc",
            }
        },
        "net": {"network": "regnet_y_1_6g", "tag": "il-common"},
    },
)
registry.register_weights(
    "regnet_y_1_6g_eu-common",
    {
        "url": "https://huggingface.co/birder-project/regnet_y_1_6g_eu-common/resolve/main",
        "description": "RegNet Y 1.6g model trained on the eu-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 42.2,
                "sha256": "a984384cbdd41d47ec179b24b49b3a9f097cc71d120caea27c2cce15b0499287",
            }
        },
        "net": {"network": "regnet_y_1_6g", "tag": "eu-common"},
    },
)
registry.register_weights(
    "regnet_y_8g_intermediate-eu-common",
    {
        "url": "https://huggingface.co/birder-project/regnet_y_8g_intermediate-eu-common/resolve/main",
        "description": "RegNet Y 8g model with intermediate training, then fine-tuned on the eu-common dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 148.5,
                "sha256": "96ac500880672ab4cf0dd6de92742a131c238ca49e6fb6040de543528c071147",
            }
        },
        "net": {"network": "regnet_y_8g", "tag": "intermediate-eu-common"},
    },
)



"""
RetinaNet, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/detection/retinanet.py

Paper "Focal Loss for Dense Object Detection", https://arxiv.org/abs/1708.02002
"""

# Reference license: BSD 3-Clause

import math
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import boxes as box_ops
from torchvision.ops import generalized_box_iou_loss
from torchvision.ops import sigmoid_focal_loss
from torchvision.ops.feature_pyramid_network import LastLevelP6P7

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import AnchorGenerator
from birder.net.detection.base import BackboneWithFPN
from birder.net.detection.base import BackboneWithSimpleFPN
from birder.net.detection.base import BoxCoder
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.base import Matcher
from birder.ops.soft_nms import SoftNMS


def _sum(x: list[torch.Tensor]) -> torch.Tensor:
    res = x[0]
    for i in x[1:]:
        res = res + i

    return res


class RetinaNetClassificationHead(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_anchors: int,
        num_classes: int,
        prior_probability: float = 0.01,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()

        conv = []
        for _ in range(4):
            conv.append(Conv2dNormActivation(in_channels, in_channels, norm_layer=norm_layer))

        self.conv = nn.Sequential(*conv)

        # Weights initialization
        for layer in self.conv.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.normal_(layer.weight, std=0.01)
                if layer.bias is not None:
                    nn.init.constant_(layer.bias, 0)

        self.cls_logits = nn.Conv2d(in_channels, num_anchors * num_classes, kernel_size=3, stride=1, padding=1)

        # Weights initialization
        nn.init.normal_(self.cls_logits.weight, std=0.01)
        nn.init.constant_(self.cls_logits.bias, -math.log((1 - prior_probability) / prior_probability))

        self.num_classes = num_classes
        self.num_anchors = num_anchors
        self.BETWEEN_THRESHOLDS = Matcher.BETWEEN_THRESHOLDS  # pylint: disable=invalid-name

    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        head_outputs: dict[str, torch.Tensor],
        matched_idxs: list[torch.Tensor],
    ) -> torch.Tensor:
        losses = []

        cls_logits = head_outputs["cls_logits"]

        for targets_per_image, cls_logits_per_image, matched_idxs_per_image in zip(targets, cls_logits, matched_idxs):
            # determine only the foreground
            foreground_idxs_per_image = matched_idxs_per_image >= 0
            num_foreground = foreground_idxs_per_image.sum()

            # Create the target classification
            gt_classes_target = torch.zeros_like(cls_logits_per_image)
            gt_classes_target[
                foreground_idxs_per_image,
                targets_per_image["labels"][matched_idxs_per_image[foreground_idxs_per_image]],
            ] = 1.0

            # Find indices for which anchors should be ignored
            valid_idxs_per_image = matched_idxs_per_image != self.BETWEEN_THRESHOLDS

            # Compute the classification loss
            losses.append(
                sigmoid_focal_loss(
                    cls_logits_per_image[valid_idxs_per_image],
                    gt_classes_target[valid_idxs_per_image],
                    reduction="sum",
                )
                / max(1, num_foreground)
            )

        return _sum(losses) / len(targets)

    def forward(self, x: list[torch.Tensor]) -> torch.Tensor:
        all_cls_logits = []

        for features in x:
            cls_logits: torch.Tensor = self.conv(features)
            cls_logits = self.cls_logits(cls_logits)

            # Permute classification output from (N, A * K, H, W) to (N, HWA, K).
            (N, _, H, W) = cls_logits.shape
            cls_logits = cls_logits.view(N, -1, self.num_classes, H, W)
            cls_logits = cls_logits.permute(0, 3, 4, 1, 2)
            cls_logits = cls_logits.reshape(N, -1, self.num_classes)  # Size=(N, HWA, K)

            all_cls_logits.append(cls_logits)

        return torch.concat(all_cls_logits, dim=1)


class RetinaNetRegressionHead(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_anchors: int,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        giou_loss: bool = True,
    ) -> None:
        super().__init__()
        self.giou_loss = giou_loss

        conv = []
        for _ in range(4):
            conv.append(
                Conv2dNormActivation(
                    in_channels, in_channels, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), norm_layer=norm_layer
                )
            )

        self.conv = nn.Sequential(*conv)
        self.bbox_reg = nn.Conv2d(in_channels, num_anchors * 4, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1))

        # Weights initialization
        for layer in self.conv.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.normal_(layer.weight, std=0.01)
                if layer.bias is not None:
                    nn.init.zeros_(layer.bias)

        self.box_coder = BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))

    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        head_outputs: dict[str, torch.Tensor],
        anchors: list[torch.Tensor],
        matched_idxs: list[torch.Tensor],
    ) -> torch.Tensor:
        losses = []
        bbox_regression = head_outputs["bbox_regression"]

        for targets_per_image, bbox_regression_per_image, anchors_per_image, matched_idxs_per_image in zip(
            targets, bbox_regression, anchors, matched_idxs
        ):
            # Determine only the foreground indices, ignore the rest
            foreground_idxs_per_image = torch.where(matched_idxs_per_image >= 0)[0]
            num_foreground = foreground_idxs_per_image.numel()

            # Select only the foreground boxes
            matched_gt_boxes_per_image = targets_per_image["boxes"][matched_idxs_per_image[foreground_idxs_per_image]]
            bbox_regression_per_image = bbox_regression_per_image[foreground_idxs_per_image, :]
            anchors_per_image = anchors_per_image[foreground_idxs_per_image, :]

            # Compute the loss
            if self.giou_loss is True:
                bbox_per_image = self.box_coder.decode_single(bbox_regression_per_image, anchors_per_image)
                loss = generalized_box_iou_loss(bbox_per_image, matched_gt_boxes_per_image, reduction="sum", eps=1e-7)
            else:
                target_regression = self.box_coder.encode_single(matched_gt_boxes_per_image, anchors_per_image)
                loss = F.l1_loss(bbox_regression_per_image, target_regression, reduction="sum") / max(1, num_foreground)

            losses.append(loss / max(1, num_foreground))

        return _sum(losses) / max(1, len(targets))

    def forward(self, x: list[torch.Tensor]) -> torch.Tensor:
        all_bbox_regression = []
        for features in x:
            bbox_regression: torch.Tensor = self.conv(features)
            bbox_regression = self.bbox_reg(bbox_regression)

            # Permute bbox regression output from (N, 4 * A, H, W) to (N, HWA, 4).
            (N, _, H, W) = bbox_regression.size()
            bbox_regression = bbox_regression.view(N, -1, 4, H, W)
            bbox_regression = bbox_regression.permute(0, 3, 4, 1, 2)
            bbox_regression = bbox_regression.reshape(N, -1, 4)  # Size=(N, HWA, 4)

            all_bbox_regression.append(bbox_regression)

        return torch.concat(all_bbox_regression, dim=1)


class RetinaNetHead(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_anchors: int,
        num_classes: int,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        giou_loss: bool = True,
    ) -> None:
        super().__init__()
        self.norm_layer = norm_layer
        self.classification_head = RetinaNetClassificationHead(
            in_channels, num_anchors, num_classes, norm_layer=norm_layer
        )
        self.regression_head = RetinaNetRegressionHead(
            in_channels, num_anchors, norm_layer=norm_layer, giou_loss=giou_loss
        )

    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        head_outputs: dict[str, torch.Tensor],
        anchors: list[torch.Tensor],
        matched_idxs: list[torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        return {
            "classification": self.classification_head.compute_loss(targets, head_outputs, matched_idxs),
            "bbox_regression": self.regression_head.compute_loss(targets, head_outputs, anchors, matched_idxs),
        }

    def forward(self, x: list[torch.Tensor]) -> dict[str, torch.Tensor]:
        return {
            "cls_logits": self.classification_head(x),
            "bbox_regression": self.regression_head(x),
        }


class RetinaNet(DetectionBaseNet):
    default_size = (640, 640)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        self.num_classes = self.num_classes - 1

        fg_iou_thresh = 0.5
        bg_iou_thresh = 0.4
        score_thresh = 0.05
        nms_thresh = 0.5
        detections_per_img = 300
        topk_candidates = 1000
        giou_loss = True
        fpn_width: int = self.config["fpn_width"]
        feature_pyramid_type: str = self.config["feature_pyramid_type"]
        soft_nms: bool = self.config.get("soft_nms", False)

        self.soft_nms = None
        if soft_nms is True:
            self.soft_nms = SoftNMS()

        if feature_pyramid_type == "fpn":
            feature_pyramid: Callable[..., nn.Module] = BackboneWithFPN
            num_anchor_sizes = len(self.backbone.return_stages) + 2
        elif feature_pyramid_type == "sfp":
            feature_pyramid = partial(BackboneWithSimpleFPN, num_stages=3)
            num_anchor_sizes = 3 + 2
        else:
            raise ValueError(f"Unknown feature_pyramid_type '{feature_pyramid_type}'")

        # Skip stage1 because it generates too many anchors (according to their paper)
        self.backbone.return_channels = self.backbone.return_channels[-3:]
        self.backbone.return_stages = self.backbone.return_stages[-3:]
        self.backbone_with_fpn = feature_pyramid(
            self.backbone,
            fpn_width,
            extra_blocks=LastLevelP6P7(self.backbone.return_channels[-1], 256),
        )

        anchor_sizes = [[x, int(x * 2 ** (1.0 / 3)), int(x * 2 ** (2.0 / 3))] for x in [32, 64, 128, 256, 512]]
        anchor_sizes = anchor_sizes[-num_anchor_sizes:]
        aspect_ratios = [[0.5, 1.0, 2.0]] * len(anchor_sizes)
        self.anchor_generator = AnchorGenerator(anchor_sizes, aspect_ratios)

        self.head = RetinaNetHead(
            self.backbone_with_fpn.out_channels,
            self.anchor_generator.num_anchors_per_location()[0],
            self.num_classes,
            norm_layer=partial(nn.GroupNorm, 32),
            giou_loss=giou_loss,
        )
        self.proposal_matcher = Matcher(fg_iou_thresh, bg_iou_thresh, allow_low_quality_matches=True)
        self.box_coder = BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))

        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img
        self.topk_candidates = topk_candidates

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes

        norm_layer = self.head.norm_layer
        self.head.classification_head = RetinaNetClassificationHead(
            self.backbone_with_fpn.out_channels,
            self.anchor_generator.num_anchors_per_location()[0],
            self.num_classes,
            norm_layer=norm_layer,
        )

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.head.classification_head.parameters():
                param.requires_grad = True

    @torch.jit.unused  # type: ignore[misc]
    @torch.compiler.disable()  # type: ignore[misc]
    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        head_outputs: dict[str, torch.Tensor],
        anchors: list[torch.Tensor],
    ) -> dict[str, torch.Tensor]:
        matched_idxs = []
        for idx, (anchors_per_image, targets_per_image) in enumerate(zip(anchors, targets)):
            if targets_per_image["boxes"].numel() == 0:
                matched_idxs.append(
                    torch.full((anchors_per_image.size(0),), -1, dtype=torch.int64, device=anchors_per_image.device)
                )
                continue

            match_quality_matrix = box_ops.box_iou(targets_per_image["boxes"], anchors_per_image)
            matched_idxs.append(self.proposal_matcher(match_quality_matrix))
            targets[idx]["labels"] = targets_per_image["labels"] - 1  # No background

        return self.head.compute_loss(targets, head_outputs, anchors, matched_idxs)

    # pylint: disable=too-many-locals
    def postprocess_detections(
        self,
        head_outputs: dict[str, list[torch.Tensor]],
        anchors: list[list[torch.Tensor]],
        image_shapes: list[tuple[int, int]],
    ) -> list[dict[str, torch.Tensor]]:
        class_logits = head_outputs["cls_logits"]
        box_regression = head_outputs["bbox_regression"]

        num_images = len(image_shapes)

        detections: list[dict[str, torch.Tensor]] = []
        for index in range(num_images):
            box_regression_per_image = [br[index] for br in box_regression]
            logits_per_image = [cl[index] for cl in class_logits]
            anchors_per_image = anchors[index]
            image_shape = image_shapes[index]

            image_boxes_list = []
            image_scores_list = []
            image_labels_list = []
            for box_regression_per_level, logits_per_level, anchors_per_level in zip(
                box_regression_per_image, logits_per_image, anchors_per_image
            ):
                num_classes = logits_per_level.shape[-1]

                # Remove low scoring boxes
                scores_per_level = torch.sigmoid(logits_per_level).flatten()
                keep_idxs = scores_per_level > self.score_thresh
                scores_per_level = scores_per_level[keep_idxs]
                topk_idxs = torch.where(keep_idxs)[0]

                # Keep only topk scoring predictions
                num_topk = min(self.topk_candidates, int(topk_idxs.size(0)))
                (scores_per_level, idxs) = scores_per_level.topk(num_topk)
                topk_idxs = topk_idxs[idxs]

                anchor_idxs = torch.div(topk_idxs, num_classes, rounding_mode="floor")
                labels_per_level = topk_idxs % num_classes
                labels_per_level += 1  # Background offset

                boxes_per_level = self.box_coder.decode_single(
                    box_regression_per_level[anchor_idxs], anchors_per_level[anchor_idxs]
                )
                boxes_per_level = box_ops.clip_boxes_to_image(boxes_per_level, image_shape)

                image_boxes_list.append(boxes_per_level)
                image_scores_list.append(scores_per_level)
                image_labels_list.append(labels_per_level)

            image_boxes = torch.concat(image_boxes_list, dim=0)
            image_scores = torch.concat(image_scores_list, dim=0)
            image_labels = torch.concat(image_labels_list, dim=0)

            # Non-maximum suppression
            if self.soft_nms is not None:
                # Actually much faster on CPU
                device = image_boxes.device
                (soft_scores, keep) = self.soft_nms(
                    image_boxes.cpu(), image_scores.cpu(), image_labels.cpu(), score_threshold=0.001
                )
                keep = keep.to(device)
                image_scores[keep] = soft_scores.to(device)
            else:
                keep = box_ops.batched_nms(image_boxes, image_scores, image_labels, self.nms_thresh)

            keep = keep[: self.detections_per_img]

            detections.append(
                {
                    "boxes": image_boxes[keep],
                    "scores": image_scores[keep],
                    "labels": image_labels[keep],
                }
            )

        return detections

    # pylint: disable=invalid-name
    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[list[int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        features: dict[str, torch.Tensor] = self.backbone_with_fpn(x)
        feature_list = list(features.values())
        head_outputs = self.head(feature_list)
        anchors = self.anchor_generator(images, feature_list)

        losses: dict[str, torch.Tensor] = {}
        detections: list[dict[str, torch.Tensor]] = []
        if self.training is True:
            assert targets is not None, "targets should not be none when in training mode"
            losses = self.compute_loss(targets, head_outputs, anchors)

        else:
            # Recover level sizes
            num_anchors_per_level = [x.size(2) * x.size(3) for x in feature_list]
            HW = 0
            for v in num_anchors_per_level:
                HW += v

            HWA = head_outputs["cls_logits"].size(1)
            A = HWA // HW
            num_anchors_per_level = [hw * A for hw in num_anchors_per_level]

            # Split outputs per level
            split_head_outputs: dict[str, list[torch.Tensor]] = {}
            for k in head_outputs:
                split_head_outputs[k] = list(head_outputs[k].split(num_anchors_per_level, dim=1))

            split_anchors = [list(a.split(num_anchors_per_level)) for a in anchors]

            # Compute the detections
            detections = self.postprocess_detections(split_head_outputs, split_anchors, images.image_sizes)

        return (detections, losses)


registry.register_model_config("retinanet", RetinaNet, config={"fpn_width": 256, "feature_pyramid_type": "fpn"})
registry.register_model_config("retinanet_sfp", RetinaNet, config={"fpn_width": 256, "feature_pyramid_type": "sfp"})



"""
ConvNeXt v1, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/convnext.py

Paper "A ConvNet for the 2020s", https://arxiv.org/abs/2201.03545
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class ConvNeXtBlock(nn.Module):
    def __init__(
        self,
        channels: int,
        layer_scale: float,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()
        self.block = nn.Sequential(
            nn.Conv2d(
                channels,
                channels,
                kernel_size=(7, 7),
                stride=(1, 1),
                padding=(3, 3),
                groups=channels,
                bias=True,
            ),
            Permute([0, 2, 3, 1]),
            nn.LayerNorm(channels, eps=1e-6),
            nn.Linear(channels, 4 * channels),  # Same as 1x1 conv
            nn.GELU(),
            nn.Linear(4 * channels, channels),  # Same as 1x1 conv
            Permute([0, 3, 1, 2]),
        )
        self.layer_scale = nn.Parameter(torch.ones(channels, 1, 1) * layer_scale, requires_grad=True)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.layer_scale * self.block(x)
        x = self.stochastic_depth(x)
        x += identity

        return x


# pylint: disable=invalid-name
class ConvNeXt_v1(DetectorBackbone):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        layer_scale = 1e-6
        in_channels: list[int] = self.config["in_channels"]
        num_layers: list[int] = self.config["num_layers"]
        drop_path_rate: float = self.config["drop_path_rate"]
        out_channels = in_channels[1:] + [-1]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            in_channels[0],
            kernel_size=(4, 4),
            stride=(4, 4),
            padding=(0, 0),
            bias=True,
            norm_layer=LayerNorm2d,
            activation_layer=None,
        )

        total_stage_blocks = sum(num_layers)
        stage_block_id = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        layers = []
        for idx, (i, out, n) in enumerate(zip(in_channels, out_channels, num_layers)):
            # Bottlenecks
            for _ in range(n):
                # Adjust stochastic depth probability based on the depth of the stage block
                sd_prob = drop_path_rate * stage_block_id / (total_stage_blocks - 1.0)
                layers.append(ConvNeXtBlock(i, layer_scale, sd_prob))
                stage_block_id += 1

            stages[f"stage{idx+1}"] = nn.Sequential(*layers)
            return_channels.append(i)
            layers = []

            # Down sampling
            if out != -1:
                layers.append(
                    nn.Sequential(
                        LayerNorm2d(i, eps=1e-6),
                        nn.Conv2d(i, out, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0), bias=True),
                    )
                )

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(in_channels[-1], eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = in_channels[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "convnext_v1_tiny",
    ConvNeXt_v1,
    config={"in_channels": [96, 192, 384, 768], "num_layers": [3, 3, 9, 3], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "convnext_v1_small",
    ConvNeXt_v1,
    config={"in_channels": [96, 192, 384, 768], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.4},
)
registry.register_model_config(
    "convnext_v1_base",
    ConvNeXt_v1,
    config={"in_channels": [128, 256, 512, 1024], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.5},
)
registry.register_model_config(
    "convnext_v1_large",
    ConvNeXt_v1,
    config={"in_channels": [192, 384, 768, 1536], "num_layers": [3, 3, 27, 3], "drop_path_rate": 0.5},
)



"""
DenseNet, adapted from
https://github.com/apache/mxnet/blob/master/python/mxnet/gluon/model_zoo/vision/densenet.py
and
https://github.com/pytorch/vision/blob/main/torchvision/models/densenet.py

Paper "Densely Connected Convolutional Networks", https://arxiv.org/abs/1608.06993
"""

# Reference license: Apache-2.0 and BSD 3-Clause

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class DenseBlock(nn.Module):
    def __init__(self, in_channels: int, num_layers: int, growth_rate: int) -> None:
        super().__init__()
        self.layers = nn.ModuleList()
        for i in range(num_layers):
            self.layers.append(
                nn.Sequential(
                    nn.BatchNorm2d(in_channels + i * growth_rate),
                    nn.ReLU(inplace=True),
                    Conv2dNormActivation(
                        in_channels + i * growth_rate,
                        4 * growth_rate,
                        kernel_size=(1, 1),
                        stride=(1, 1),
                        padding=(0, 0),
                        bias=False,
                    ),
                    nn.Conv2d(
                        4 * growth_rate,
                        growth_rate,
                        kernel_size=(3, 3),
                        stride=(1, 1),
                        padding=(1, 1),
                        bias=False,
                    ),
                )
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for layer in self.layers:
            dense_branch = layer(x)
            x = torch.concat((x, dense_branch), dim=1)

        return x


class TransitionBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.block = nn.Sequential(
            nn.BatchNorm2d(in_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.block(x)
        return x


class DenseNet(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        growth_rate: int = self.config["growth_rate"]
        num_init_features: int = self.config["num_init_features"]
        layer_list: list[int] = self.config["layer_list"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                num_init_features,
                kernel_size=(7, 7),
                stride=(2, 2),
                padding=(3, 3),
                bias=False,
            ),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        # Add dense blocks
        num_features = num_init_features
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        layers = []
        for i, num_layers in enumerate(layer_list):

            layers.append(DenseBlock(num_features, num_layers=num_layers, growth_rate=growth_rate))
            num_features = num_features + (num_layers * growth_rate)

            stages[f"stage{i+1}"] = nn.Sequential(*layers)
            return_channels.append(num_features)
            layers = []

            if i != len(layer_list) - 1:
                layers.append(TransitionBlock(num_features, num_features // 2))
                num_features = num_features // 2

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "densenet_121", DenseNet, config={"growth_rate": 32, "num_init_features": 64, "layer_list": [6, 12, 24, 16]}
)
registry.register_model_config(
    "densenet_161", DenseNet, config={"growth_rate": 48, "num_init_features": 96, "layer_list": [6, 12, 36, 24]}
)
registry.register_model_config(
    "densenet_169", DenseNet, config={"growth_rate": 32, "num_init_features": 64, "layer_list": [6, 12, 32, 32]}
)
registry.register_model_config(
    "densenet_201", DenseNet, config={"growth_rate": 32, "num_init_features": 64, "layer_list": [6, 12, 48, 32]}
)



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import Optional

import torch
from torchvision.datasets import CocoDetection
from torchvision.datasets import wrap_dataset_for_transforms_v2
from torchvision.transforms.v2 import functional as F

logger = logging.getLogger(__name__)


def _remove_images_without_annotations(dataset: CocoDetection, ignore_list: list[str]) -> CocoDetection:
    def _has_only_empty_bbox(anno: list[dict[str, Any]]) -> bool:
        return all(any(o <= 1 for o in obj["bbox"][2:]) for obj in anno)

    def _has_valid_annotation(anno: list[dict[str, Any]]) -> bool:
        # If it's empty, there is no annotation
        if len(anno) == 0:
            return False

        # If all boxes have close to zero area, there is no annotation
        if _has_only_empty_bbox(anno):
            return False

        return True

    ids = []
    for ds_idx, img_id in enumerate(dataset.ids):
        img_path = dataset.coco.loadImgs(img_id)[0]["file_name"]
        if img_path in ignore_list:
            logger.debug(f"Ignoring {img_path}")
            continue

        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        if _has_valid_annotation(anno):
            ids.append(ds_idx)

    return torch.utils.data.Subset(dataset, ids)


def _convert_to_binary_annotations(dataset: CocoDetection) -> CocoDetection:
    for img_id in dataset.ids:
        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        for obj in anno:
            obj["category_id"] = 1

    return dataset


class CocoBase(torch.utils.data.Dataset):
    def __init__(
        self, root: str | Path, ann_file: str, transforms: Optional[Callable[..., torch.Tensor]] = None
    ) -> None:
        super().__init__()
        dataset = CocoDetection(root, ann_file, transforms=transforms)
        self.class_to_idx = {cat["name"]: cat["id"] for cat in dataset.coco.cats.values()}

        # The transforms v2 wrapper causes open files count to "leak"
        # It seems due to the Pythonic COCO objects, maybe related to
        # https://ppwwyyxx.com/blog/2022/Demystify-RAM-Usage-in-Multiprocess-DataLoader/
        self.dataset = wrap_dataset_for_transforms_v2(dataset)

    def __getitem__(self, index: int) -> tuple[Any, ...]:
        raise NotImplementedError

    def __len__(self) -> int:
        return len(self.dataset)

    def __repr__(self) -> str:
        return repr(self.dataset)

    def remove_images_without_annotations(self, ignore_list: list[str]) -> None:
        self.dataset = _remove_images_without_annotations(self.dataset, ignore_list)

    def convert_to_binary_annotations(self) -> None:
        self.dataset = _convert_to_binary_annotations(self.dataset)
        self.class_to_idx = {"Object": 1}


class CocoTraining(CocoBase):
    def __getitem__(self, index: int) -> tuple[torch.Tensor, Any]:
        (sample, labels) = self.dataset[index]
        return (sample, labels)


class CocoInference(CocoBase):
    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any, list[int]]:
        coco_id = self.dataset.ids[index]
        path = self.dataset.coco.loadImgs(coco_id)[0]["file_name"]
        (sample, labels) = self.dataset[index]

        return (path, sample, labels, F.get_size(sample))
